/* Copyright (C) 2007-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Igor Kulakov, Sergey Gorbunov [committer], Andrey Lebedev, Sergei Zharko */

#ifndef CaMaterialMap_h
#define CaMaterialMap_h 1

#include <boost/serialization/vector.hpp>

#include <iomanip>
#include <sstream>
#include <string>
#include <vector>

#include "CaSimd.h"
#include "CaUtils.h"

namespace cbm::algo::ca
{
  /// \class MaterialMap
  /// \brief A map of station thickness in units of radiation length (X0) to the specific point in XY plane
  class MaterialMap {
  public:
    /// \brief Default constructor
    MaterialMap() = default;

    /// \brief Copy constructor
    MaterialMap(const MaterialMap& other) = default;

    /// \brief Copy assignment operator
    MaterialMap& operator=(const MaterialMap& other) = default;

    /// \brief Move constructor
    MaterialMap(MaterialMap&& other) noexcept;

    /// \brief Move assignment operator
    MaterialMap& operator=(MaterialMap&& other) noexcept;

    /// \brief Destructor
    ~MaterialMap() noexcept = default;

    /// \brief Gets number of bins (rows or columns) of the material table
    int GetNbins() const { return fNbins; }

    /// \brief Gets radius in cm of the material table
    float GetXYmax() const { return fXYmax; }

    /// \brief Gets reference Z of the material in cm
    float GetZref() const { return fZref; }

    /// \brief Gets minimal Z of the collected material in cm
    float GetZmin() const { return fZmin; }

    /// \brief Gets maximal Z of the collected material in cm
    float GetZmax() const { return fZmax; }

    /// \brief Gets value of X/X0 in a given cell of the material table by the indeces of the row and column
    /// \param iBinX  Index of table column
    /// \param iBinY  Index of table row
    float GetRadThickBin(int iBinX, int iBinY) const { return fTable[iBinX + fNbins * iBinY]; }

    /// \brief Gets value of X/X0 in a given cell of the material table by the index of the bin
    /// \param iBin  Index of the bin in 2d table
    float GetRadThickBin(int iBin) const { return fTable[iBin]; }

    /// \brief Gets material thickness in units of X0 in (x,y) point of the station
    /// \param x  X coordinate of the point [cm]
    /// \param y  Y coordinate of the point [cm]
    float GetRadThickScal(float x, float y) const;

    /// \brief Gets material thickness in units of X0 in (x,y) point of the station
    /// cbm::algo::ca::fvec type can be float, that is why "Vec" and "Scal" specifications
    /// \param x  X coordinate of the point [cm] (SIMDized vector)
    /// \param y  Y coordinate of the point [cm] (SIMDized veotor)
    fvec GetRadThickVec(fvec x, fvec y) const;

    /// \brief Checks, if the fields are NaN
    bool IsNaN() const
    {
      return utils::IsUndefined(fNbins) || utils::IsUndefined(fXYmax) || utils::IsUndefined(fFactor)
             || utils::IsUndefined(fZref) || utils::IsUndefined(fZmin) || utils::IsUndefined(fZmax);
    }

    /// \brief Verifies class invariant consistency
    void CheckConsistency() const;

    /// \brief Sets value of material thickness in units of X0 for a given cell of the material table
    /// \param iBinX      Index of table column
    /// \param iBinY      Index of table row
    /// \param thickness  Thickness of the material in units of X0
    /// \note  Indices of rows and columns in the table runs from 0 to nBins-1 inclusively, where nBins is the number
    ///        both of rows and columns. One should be careful while reading and storing the table from ROOT-file,
    ///        because iBinX = 0 and iBinY = 0 in the TH1::SetBinContent method of usually defines the underflow bin.
    void SetRadThickBin(int iBinX, int iBinY, float thickness) { fTable[iBinX + fNbins * iBinY] = thickness; }

    /// \brief Sets properties of the material table -- number of rows or columnts and the size of station in XY plane
    /// \param nBins  Number of rows or columns
    /// \param xyMax  Size of station in x and y dimensions [cm]
    void Initialize(int nBins, float xyMax, float zRef, float zMin, float zMax);

    /// \brief Swap method
    void Swap(MaterialMap& other) noexcept;

    /// \brief Get bin index for (x,y). Returns -1 when outside of the map
    int GetBin(float x, float y) const;

  private:
    int fNbins    = constants::Undef<int>;    ///< Number of rows (== N columns) in the material budget table
    float fXYmax  = constants::Undef<float>;  ///< Size of the station in x and y dimensions [cm]
    float fFactor = constants::Undef<float>;  ///< Util. var. for the conversion of point coordinates to row/column id
    float fZref   = constants::Undef<float>;  ///< Reference Z of the collected material [cm]
    float fZmin   = constants::Undef<float>;  ///< Minimal Z of the collected material [cm]
    float fZmax   = constants::Undef<float>;  ///< Minimal Z of the collected material [cm]
    std::vector<float> fTable {};             ///< Material budget table

    /// \brief Serialization function
    friend class boost::serialization::access;
    template<class Archive>
    void serialize(Archive& ar, const unsigned int)
    {
      ar& fNbins;
      ar& fXYmax;
      ar& fFactor;
      ar& fZref;
      ar& fZmin;
      ar& fZmax;
      ar& fTable;
    }
  } _fvecalignment;

}  // namespace cbm::algo::ca

#endif
