/* Copyright (C) 2022-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/// \file   CaUtils.h
/// \brief  Compile-time constants definition for the CA tracking algorithm
/// \since  02.06.2022
/// \author S.Zharko <s.zharko@gsi.de>

#ifndef CaUtils_h
#define CaUtils_h 1

#include <sstream>

#include "AlgoFairloggerCompat.h"
#include "CaConstants.h"
#include "CaSimd.h"

/// Namespace contains compile-time constants definition for the CA tracking algorithm
///
namespace cbm::algo::ca::utils
{

  inline fvec iif(const fmask& m, const fvec& t, const fvec& f) { return Vc::iif(m, t, f); }
  inline fvec fabs(const fvec& v) { return Vc::abs(v); }

  template<typename T>
  inline T iif(bool b, T t, T f)
  {
    return b ? t : f;
  }


  template<typename T>
  inline T fabs(const T& v)
  {
    return std::fabs(v);
  }

  /// \brief Checks whether a variable of a particular type defined
  /// \param val Value to be checked
  template<typename T>
  inline bool IsUndefined(T val)
  {
    if constexpr (std::is_same_v<T, float>) { return std::isnan(val); }
    else if constexpr (std::is_same_v<T, double>) {
      return std::isnan(val);
    }
    else if constexpr (std::is_same_v<T, fscal>) {
      return std::isnan(val);
    }
    else if constexpr (std::is_same_v<T, fvec>) {
      return isnan(val).isNotEmpty();
    }
    else {
      return val == cbm::algo::ca::constants::Undef<T>;
    }
  }

  /// \brief Checks whether a variable of a particular type is finite
  /// \param val  Value to check
  template<typename T>
  inline bool IsFinite(T val)
  {
    if constexpr (std::is_same_v<T, float>) { return std::isfinite(val); }
    else if constexpr (std::is_same_v<T, double>) {
      return std::isfinite(val);
    }
    else if constexpr (std::is_same_v<T, fscal>) {
      return std::isfinite(val);
    }
    else if constexpr (std::is_same_v<T, fvec>) {
      return isfinite(val).isFull();
    }
    else {
      return val != cbm::algo::ca::constants::Undef<T>;
    }
  }

  /// \brief Checks, if a SIMD vector horizontally equal
  /// TODO: Find this method in the VC!
  [[gnu::always_inline]] inline void CheckSimdVectorEquality(fvec v, const char* name)
  {
    bool ok = true;
    for (size_t i = 1; i < fvec::size(); i++) {
      ok = ok && (v[i] == v[0]);
    }
    if (!ok) {
      std::stringstream msg;
      msg << name << " SIMD vector is inconsistent, not all of the words are equal each other: " << v;
      throw std::logic_error(msg.str());
    }
  }

  /// \brief  Checks, if a particular value lies within selected limits
  /// \param  name   Name of parameters
  /// \param  value  Value of parameter
  /// \param  lLimit Lower limit of parameter
  /// \param  uLimit Upper limit of parameter
  template<typename T>
  bool CheckValueLimits(const std::string& name, T value, T lLimit, T uLimit)
  {
    if (value < lLimit || value > uLimit) {
      LOG(error) << "parameter\033[1;32m" << name << "\033[0m (" << value << ") runs out of range: [" << lLimit << ','
                 << uLimit << ']';
      return false;
    }
    return true;
  }

  /// \struct EnumClassHash
  /// \brief  Hash for unordered_map with enum class keys
  struct EnumClassHash {
    template<typename T>
    int operator()(T t) const
    {
      return static_cast<int>(t);
    }
  };
}  // namespace cbm::algo::ca::utils


#endif  // CaUtils_h
